{------------------------------------------------------------------------------}
{
{                                  SCB.PAS
{
{    Script Beautifier - Conversor de Scripts
{
{    Sistema:    DosVox
{    Mdulo:     Interpretador ScriptVox
{    Autor:      Oswaldo Vernet
{    Data:       21/08/2015
{    Alteraes: 29/03/2016
{
{------------------------------------------------------------------------------}

program scb;

uses
    sysUtils, classes, windows, dvwin, dvcrt, interp;

{--------------------------------------------------------}
{            Variveis globais
{--------------------------------------------------------}

label
    usage;

var
    quiet      : boolean;
    script     : string;
    nconverted : integer;

{--------------------------------------------------------}
{            Escreve ou fala as mensagens
{--------------------------------------------------------}

procedure issue (msg : string);
var
    sameline : boolean;
begin
    sameline := msg[length(msg)] in ['?', ':'];

    if quiet then
        if sameline then
            write (msg + ' ')
        else
            writeln (msg)
    else
        if sameline then
            sintWrite (msg + ' ')
        else
            sintWriteln (msg)
end;

{--------------------------------------------------------}
{            Converte um script
{--------------------------------------------------------}

procedure convert (src, dst : string);
label
    finish;
var
    s                        : integer;
    answer, nameSrc, nameDst : string;
begin
    nameSrc := ExtractFileName (src);
    nameDst := ExtractFileName (dst);

    writeln ('Convertendo ' + nameSrc + ' para ' + nameDst);

    s := loadScript (src);

    if s < 0 then
    begin
        issue ('Script ' + nameSrc + ' no encontrado');
        exit
    end;

    if FileExists (dst) then
    begin
        issue ('O script ' + nameDst + ' j existe. Sobrescreve?');

        if quiet then
            readln (answer)
        else
            sintReadln (answer);

        answer := LowerCase (answer);

        if not (answer[1] in ['s', 'y']) then
            goto finish
    end;

    beautifyScript (s, dst);
    INC (nconverted);

finish:
    freeScript (s)
end;

{--------------------------------------------------------}
{            Percorre uma subrvore
{--------------------------------------------------------}

procedure explore (src, dst : string);
var
    s : TSearchRec;
begin
    if not DirectoryExists (dst) then
        mkdir (dst);

    if FindFirst (src + '\*.*', faDirectory, s) = 0 then
    begin
        repeat
            if (s.name <> '.') and (s.name <> '..') then
            begin
                if (s.Attr and faDirectory) <> 0 then
                    explore (src + '\' + s.name, dst + '\' + s.name)
                else if copy (s.name, length (s.name) - 3, 4) = '.cmd' then
                    convert (src + '\' + s.name, dst + '\' + copy (s.name, 1, length (s.name) - 4) + '.pro')
            end
        until FindNext (s) <> 0
    end
end;

{--------------------------------------------------------}
{            Realiza o processamento
{--------------------------------------------------------}

procedure process (path : string);
var
    answer   : string;
    dir      : boolean;
    src, dst : string;
begin
    src := ''; dst := ''; dir := false;

    path := ExpandFileName (path);

    if DirectoryExists (path) then
    begin
        dir := true;
        src := path;
        dst := path + '-PRO'
    end
    else if FileExists (path) then
    begin
        if copy (path, length (path) - 3, 4) = '.cmd' then
        begin
            src := path;
            dst := copy (path, 1, length (path) - 4) + '.pro'
        end
        else begin
            issue ('O arquivo dado NO tem extenso .cmd')
        end
    end
    else if copy (path, length (path) - 4, 4) = '.cmd' then
    begin
        issue ('O arquivo ' + path +  ' NO existe')
    end
    else if FileExists (path + '.cmd') then
    begin
        src := path + '.cmd';
        dst := path + '.pro'
    end
    else begin
        issue ('O arquivo ' + ExtractFileName (path) + ' NO existe');
        issue ('Fim do programa. Tecle ENTER.');
        readln;
        exit
    end;

    nconverted := 0;

    if dir then
    begin
        issue ('Diretrio fonte: '   + ExtractFileName (src));
        issue ('Diretrio destino: ' + ExtractFileName (dst));

        issue ('Realiza a converso?');

        if quiet then
            readln (answer)
        else
            sintReadln (answer);

        answer := LowerCase (answer);

        if answer[1] in ['s', 'y'] then
            explore (src, dst)
    end
    else begin
        convert (src, dst)
    end;

    if nconverted > 0 then
        issue (intToStr (nconverted) + ' arquivos convertidos. Tecle ENTER para finalizar.')
    else
        issue ('Nenhuma converso realizada. Tecle ENTER para finalizar.');
    readln
end;

{--------------------------------------------------------}
{       Programa Principal do Script Beautifier
{--------------------------------------------------------}

begin
    sintInic (0, sintAmbiente ('SCRIPVOX', 'DIRSCRIPVOX'));

    clrscr; setWindowText (crtWindow, 'SCB');

    quiet := false;

    case ParamCount of
        0:      begin
                    issue ('Nome do arquivo .CMD ou diretrio a converter:');
                    sintReadln (script);
                    process (script)
                end;
        1:      begin
                    process (ParamStr (1))
                end;
        2:      begin
                    if LowerCase (ParamStr (1)) = 'mudo' then
                    begin
                        quiet := true;
                        process (ParamStr (2))
                    end
                    else goto usage
                end
        else    goto usage
    end;

    exit;

usage:
    sintWriteln ('Programa SCB - Conversor de Scripts CMD para PRO.');
    sintWriteln ('Modo de usar: scb script ou diretrio.');
    sintWriteln ('Se no quiser ouvir as mensagens: scb MUDO script ou diretrio.');
    sintWriteln ('Tecle ENTER para fechar a janela.');
    readln
end.
